@@ -14,6 +14,21 @@
 #include "compute_gx_gy.h" //only_for_context none
 //include_file compute_gx_gy.h for_context cpu_serial opencl cuda cpu_openmp
 
+/*gpufun*/
+void get_charge_density(const double x,
+                      const double y,
+                      const double sigma_x,
+                      const double sigma_y,
+                      double* rho)
+{
+
+  // this is a PDF
+  double factor = 1 / (2*PI*sigma_x*sigma_y);
+  double exp_x = exp(-x*x/(2*sigma_x*sigma_x));
+  double exp_y = exp(-y*y/(2*sigma_y*sigma_y));
+  *rho = factor * exp_x * exp_y;  // [m^-2]
+}
+
 /*gpufun*/
 void get_transv_field_gauss_round(
     double sigma, double Delta_x, double Delta_y,
@@ -24,8 +39,10 @@ void get_transv_field_gauss_round(
   double r2, temp;
 
   r2 = (x-Delta_x)*(x-Delta_x)+(y-Delta_y)*(y-Delta_y);
-  if (r2<1e-20) temp = sqrt(r2)/(2.*PI*EPSILON_0*sigma); //linearised
-  else          temp = (1-exp(-0.5*r2/(sigma*sigma)))/(2.*PI*EPSILON_0*r2);
+  double const inv_sig2 = 1./(sigma*sigma);
+  double const r2_sig2 = r2 * inv_sig2;
+  if (r2_sig2<1e-10) temp = inv_sig2/(4.*PI*EPSILON_0); //linearised
+  else temp = inv_sig2 * (1-exp(-0.5*r2_sig2))/(2.*PI*EPSILON_0*r2_sig2);
 
   (*Ex) = temp * (x-Delta_x);
   (*Ey) = temp * (y-Delta_y);
